using GraphicsManager.Interfaces;

namespace GraphicsManager.Objects.Core;

public class ControlList
{
    internal List<RenderObjectCore> _internal = new();
    private bool Clearing = false;

    public RenderObjectCore this[int Index] => _internal[Index];
    public RenderObjectCore this[uint Index] => _internal[(int)Index];

    public IEnumerable<RenderObjectCore> Where(Func<RenderObjectCore, bool> func) => _internal.Where(func);
    
    public bool Contains(RenderObjectCore Control) => _internal.Contains(Control);
    
    public int IndexOf(RenderObjectCore Control) => _internal.IndexOf(Control);

    public int Length => _internal.Count;

    public event Func<int, RenderObjectCore, Task>? ControlAdded;
    public event Func<int, RenderObjectCore, Task>? ControlAfterAdded;
    public event Func<Task>? ControlRemoved;

    public void Remove(RenderObjectCore item, bool purge = true, bool GCC = false)
    {
        _internal.Remove(item);
        if (purge)
        {
            item.Clean();
            item = null!;
        }
        if (GCC) GC.Collect();
        if (ControlRemoved is not null && !Clearing) _ = ControlRemoved.Invoke();
    }

    public void MoveControlToEnd(RenderObjectCore item)
    {
        _internal.Remove(item);
        _internal.Add(item);
    }
    
    public void Add(RenderObjectCore item)
    {
        if (ControlAdded is not null) ControlAdded.Invoke(_internal.Count, item).Wait();
        _internal.Add(item);
        if (ControlAfterAdded is not null) ControlAfterAdded.Invoke(_internal.Count-1, item).Wait();
    }
    
    public void Insert(int index, RenderObjectCore item)
    {
        if (ControlAdded is not null) ControlAdded.Invoke(index, item).Wait();
        _internal.Insert(index, item);
        if (ControlAfterAdded is not null) ControlAfterAdded.Invoke(index, item).Wait();
    }
    
    public void Clear(bool purge = true, bool GCC = false)
    {
        Clearing = true;
        for (int i = 0; i< _internal.Count; i++)
        {
            if (purge)
            {
                _internal[i].Clean();
                _internal[i] = null!;
            }
        }
        _internal.Clear();
        if (GCC) GC.Collect();
        Clearing = false;
        if (ControlRemoved is not null) _ = ControlRemoved.Invoke();
    }
}