using System.Text;
using GraphicsManager.Interfaces;
using GraphicsManager.Objects.Core.Fonts;
using GraphicsManager.Structs;
using OpenTK.Graphics.OpenGL4;
using OpenTK.Mathematics;
using OpenTK.Windowing.Common;
using OpenTK.Windowing.Desktop;
using OpenTK.Windowing.GraphicsLibraryFramework;

namespace GraphicsManager.Objects.Core;

public class LabelBase : RenderObjectCore
{
	public LabelBase(FontFamily fontFamily)
	{
		Font = FontInteraction.Load(fontFamily);
	}
	
	public LabelBase(FontInteraction interaction)
	{
		Font = interaction;
	}

	protected Vector2i? msize = null;
	public virtual Vector2i? MaxSize
	{
		get => msize;
		set
		{
			msize = value;
			StringBuilder sb = new();
			float max_x = 0, lines = 1, char_x = 0F;
			for (int i = 0; i < Text.Length; i++)
			{
				char c;
				if (PasswordChar is null)
					c = Text[i];
				else
					c = PasswordChar.Value;
				bool n = (c == '\n');

                if (n)
                {
	                lines++;
	                sb.Append(c);
                	char_x = 0f;
                }
                else
                {
	                Character ch;
		            if (Window is not null) ch = Texture.GetChar(Font, c, Window.Context);
		            else ch = Texture.GetChar(Font, c);
					if (i > 0 && text[i-1] == ' ')
					{
					    int addc = 0;
					    float word_char_x = char_x;
					    while (MaxSize is not null)
					    {
					        if (addc + i == Text.Length) break;
					        if (text[addc + i] == ' ') break;
					        Character ch2;
					        if (Window is not null) ch2 = Texture.GetChar(Font, text[addc + i], Window.Context);
					        else ch2 = Texture.GetChar(Font, c);
					        word_char_x += (ch2.Advance >> 6) * Scale;
					        if (word_char_x > MaxSize.Value.X)
					        {
					            char_x = 0f;
					            lines++;
					            sb.Append('\n');
					            break;
					        }
					        addc++;
					    }
					}
					sb.Append(c);
                	float w = ch.Size.X * Scale;
                	float xrel = char_x + ch.Bearing.X * Scale;
                	char_x += (ch.Advance >> 6) * Scale;
	                if ((xrel + w) >= max_x) max_x = (xrel + w);
                }
			}
			text_Calculated = sb.ToString();
			Size = new((int)max_x, (int)(lines * LineHeight) + (int)(lines * Font.ExtraLinePixels));
		}
	}
	protected char? pc;
	public char? PasswordChar
	{
		get => pc;
		set
		{
			pc = value; 
			if (Parent is not null) Parent.TryDraw();
		}
	}

	public static readonly Dictionary<IGLFWGraphicsContext, Dictionary<FontInteraction, Dictionary<uint, Character>>> _characters = new();
	protected string text = string.Empty;
	protected string text_Calculated = string.Empty;
	public int VAO { get; protected set; }
	
	public int VBO { get; protected set; }
	public virtual Vector2 DIR { get; set; } = new Vector2(1f, 0f);
	
	public virtual Vector2i GetSizeOfChar(int Index)
	{
		float mx = 0, my = 0;
		char character;
		if (PasswordChar is null)
			character = Text[Index];
		else
			character = PasswordChar.Value;
				
				
		if (character == '\n')
		{
			mx = 0f;
			my += Font.PixelHeight * Scale;
		}
		else
		{
			Character cha;
			if (Window is not null) cha = Texture.GetChar(Font, character, Window.Context);
			else cha = Texture.GetChar(Font, character);
			mx = cha.Size.X * Scale;
			my = cha.Size.Y * Scale;
		}
		return new((int)mx, (int)my);
	}
	
	public virtual Vector2i GetBearingOfChar(int Index)
	{
		float mx = 0, my = 0;
		char character;
		if (PasswordChar is null)
			character = Text[Index];
		else
			character = PasswordChar.Value;
				
		if (character == '\n')
		{
			mx = 0f;
			my += Font.PixelHeight * Scale;
		}
		else
		{
			Character cha;
			if (Window is not null) cha = Texture.GetChar(Font, character, Window.Context);
			else cha = Texture.GetChar(Font, character);
			mx = cha.Bearing.X * Scale;
			my = cha.Bearing.Y * Scale;
		}
		return new((int)mx, (int)my);
	}

	
	public virtual string Text
	{
		get => text;
		set
		{
			if (value is null) value = string.Empty;
			text = value;
			StringBuilder sb = new();
			text_Calculated = string.Empty;
			float max_x = 0, lines = 1, char_x = 0F;
			for (int i = 0; i < Text.Length; i++)
			{
				char c;
				if (PasswordChar is null)
					c = Text[i];
				else
					c = PasswordChar.Value;
				bool n = (c == '\n');

				if (n)
				{
					lines++;
					char_x = 0f;
					sb.Append(c);
				}
				else
				{
					Character ch;
					if (Window is not null) ch = Texture.GetChar(Font, c, Window.Context);
					else ch = Texture.GetChar(Font, c);
					if (MaxSize is not null && i > 0 && text[i-1] == ' ')
					{
						int addc = 0;
						float word_char_x = char_x;
						while (true)
						{
							if (addc + i == Text.Length) break;
							if (text[addc + i] == ' ') break;
							Character ch2;
							if (Window is not null) ch2 = Texture.GetChar(Font, text[addc + i], Window.Context);
							else ch2 = Texture.GetChar(Font, c);
							word_char_x += (ch2.Advance >> 6) * Scale;
							if (word_char_x > MaxSize.Value.X)
							{
								char_x = 0f;
								lines++;
								sb.Append('\n');
								break;
							}
							addc++;
						}
					}

					sb.Append(c);
					float w = ch.Size.X * Scale;
					float xrel = char_x + ch.Bearing.X * Scale;
					char_x += (ch.Advance >> 6) * Scale;
					if ((xrel + w) >= max_x) max_x = (xrel + w);
				}
			}

			text_Calculated = sb.ToString();
			Size = new((int)max_x, (int)(lines * LineHeight) + (int)(lines * Font.ExtraLinePixels));
			if (Loaded)
			{
				
				if (Window is not null && Window.CanControleUpdate)
				{
					if (!Window.Context.IsCurrent)
					{
						try
						{
							Window.Context.MakeCurrent();
							GL.PixelStore(PixelStoreParameter.UnpackAlignment, 1);
							
						}
						catch (Exception e)
						{
							Console.WriteLine(e);
						}
					}
					
					TryDraw();
				}
			}
		}
	}
	public virtual uint LineHeight
	{
		get
		{
			return (uint)((Font.PixelHeight * (Font.CurrentFonts[0].Face.Height / (float)Font.CurrentFonts[0].Face.UnitsPerEM)) * Scale);
		}
	}
	public Shader Shader { get; set; } = null!;
	public FontInteraction Font { get; protected set; }
	public virtual float Scale { get; set; } = 1.0f;
	
	public virtual void SetLocation(int x, int y)
	{
		loc_.X = x;
		loc_.Y = y;
		if (Window is null || Parent is null) return;
		if (Window.CanControleUpdate && Loaded)
		{
			Parent!.TryDraw();
			if (!Window.Context.IsCurrent) Window.Context.MakeCurrent();
		}
	}

	public override void UpdateOpenGLCords(bool draw = false)
	{
		
	}

	public virtual Vector2i GetCharLocation(int index)
	{
		float max_x = 0, lines = 0, char_x = 0F;
		for (int i = 0; i < index; i++)
		{
			char c;
			if (PasswordChar is null)
				c = text_Calculated[i];
			else
				c = PasswordChar.Value;
			bool n = (c == '\n');

			if (n)
			{
				lines++;
				char_x = 0f;
			}
			else
			{
				Character ch;
				if (Window is not null) ch = Texture.GetChar(Font, c, Window.Context);
				else ch = Texture.GetChar(Font, c);
				float w = ch.Size.X * Scale;
				float xrel = char_x + ch.Bearing.X * Scale;
				char_x += (ch.Advance >> 6) * Scale;
				max_x = (xrel + w);
			}
		}
		return new((int)char_x, (int)(lines * LineHeight) + (int)(lines * Font.ExtraLinePixels));
	}
	
	public virtual void SendKeyEvent(KeyboardKeyEventArgs KeyArgs)
	{
		
	}
	
	public virtual void SendClipEvent(string ClipString)
	{
		
	}

	internal override void Clean()
	{
		Tuple<int, int, int> tup = GlobalBuffers[Window!.Context];
		if (tup.Item3 - 1 == 0)
		{
			//Broken, I may fix latter
			
			//if (!Window.Context.IsCurrent) Window.Context.MakeCurrent();
			//GL.DeleteBuffer(VBO);
			//GL.DeleteVertexArray(VAO);
			//GL.BindBuffer(BufferTarget.ArrayBuffer, 0);
			//GlobalBuffers.Remove(Window!.Context);
		}
		else
		{
			GlobalBuffers[Window!.Context] = new(tup.Item1, tup.Item2, tup.Item3 - 1);
		}
		Size = new(0, 0);
		Loaded = false;
		Visible = false;
	}
	
	
	public override void TryDraw()
	{
		if (Parent is not null) Parent.TryDraw(1);
	}
	
	

	public virtual void NotifiNotVisible()
	{
		IsVisible = false;
	}
	
	public override void DrawSelf()
	{
		if (Visible && Loaded && this.Font is not null)
		{
			if (!Window!.Context.IsCurrent) Window.Context.MakeCurrent();
			Shader.Use();
			GL.Uniform4(Shader.GetUniformLocation("textColor"), new Color4(255,255,255,255));
			GL.BlendFunc(BlendingFactor.SrcAlpha, BlendingFactor.OneMinusSrcAlpha);
			GL.BlendFunc(0, BlendingFactorSrc.SrcAlpha, BlendingFactorDest.OneMinusSrcAlpha);
			Shader.SetMatrixF4("projection", Window.WindowSizeMatrix);

			GL.BindVertexArray(VAO);

			float angle_rad = (float)Math.Atan2(DIR.Y, DIR.X);
			Matrix4 rotateM = Matrix4.CreateRotationZ(angle_rad);
			Matrix4 transOriginM = Matrix4.CreateTranslation(new Vector3(loc_.X + Parent!.IntToWindow(0), loc_.Y + (Font.PixelHeight * Scale) + Parent!.IntToWindow(0, true), 0f));
			float char_x = 0.0f;
			Matrix4 staticTransform = rotateM * transOriginM;
			GL.UniformMatrix4(0, false, ref staticTransform);
			
			GL.PixelStore(PixelStoreParameter.UnpackAlignment, 1);
			GL.ActiveTexture((Shader.GetUniformLocation("u_texture") switch
			{
				0 => TextureUnit.Texture0,
				1 => TextureUnit.Texture1,
				2 => TextureUnit.Texture2,
				3 => TextureUnit.Texture3,
				4 => TextureUnit.Texture4,
				5 => TextureUnit.Texture5,
				6 => TextureUnit.Texture6,
				7 => TextureUnit.Texture7,
				8 => TextureUnit.Texture8,
				9 => TextureUnit.Texture9,
			}));
			
			float hhh = 0f;
			for (int i = 0; i < text_Calculated.Length; i++)
			{
				char c;
				if (PasswordChar is null)
					c = text_Calculated[i];
				else
					c = PasswordChar.Value;
				bool n = (c == '\n');
				if (!_characters[Window!.Context][Font].ContainsKey(c) && !n)
				{
					_ = Texture.TextureForChar(Window!.Context, Font, c, Shader);
				}
				
                if (n)
                {
                	hhh += LineHeight;
	                hhh += Font.ExtraLinePixels;
                	char_x = 0f;
                }
                else
                {
	                if (!_characters[Window!.Context][Font].ContainsKey(c)) continue;
	                Character ch = _characters[Window!.Context][Font][c];
					
                	float w = ch.Size.X * Scale;
                	float h = ch.Size.Y * Scale;
                	float xrel = char_x + ch.Bearing.X * Scale;
                	float yrel = (ch.Size.Y - ch.Bearing.Y) * Scale;
                	yrel += hhh;
                	char_x += (ch.Advance >> 6) * Scale;
	                
	                GL.Uniform2(Shader.GetUniformLocation("scale"), w, h);
	                GL.Uniform2(Shader.GetUniformLocation("offset"), xrel, yrel);

                	ch.Texture.Use();

                	GL.DrawArrays(PrimitiveType.Triangles, 0, 6);
                }
			}
        }
	}
	protected static Dictionary<IGLFWGraphicsContext, Tuple<int, int, int>> GlobalBuffers = new();


	internal override void LoadToParent(IParent window, IWindow win)
	{
		if (Loaded) return;
		if (!_characters.ContainsKey(win!.Context)) _characters.Add(win!.Context, new());
		if (!_characters[win!.Context].ContainsKey(Font)) _characters[win!.Context].Add(Font, new Dictionary<uint, Character>());
		if (Shader is null) Shader = Label.DefaultTextShader[win.Context];
		Parent = window;
		Window = win;
		Window.MouseDown += WindowOnMouseDown;
		if (!Window!.Context.IsCurrent) Window.Context.MakeCurrent();
		GL.PixelStore(PixelStoreParameter.UnpackAlignment, 1);
		GL.PixelStore(PixelStoreParameter.UnpackAlignment, 4);
		if (!GlobalBuffers.ContainsKey(win.Context))
		{
			float[] vquad =
			{
				0.0f, -1.0f,   0.0f, 0.0f,
				0.0f,  0.0f,   0.0f, 1.0f,
				1.0f,  0.0f,   1.0f, 1.0f,
				0.0f, -1.0f,   0.0f, 0.0f,
				1.0f,  0.0f,   1.0f, 1.0f,
				1.0f, -1.0f,   1.0f, 0.0f
			};
			int _VBO = GL.GenBuffer();
			GL.BindBuffer(BufferTarget.ArrayBuffer, _VBO);
			GL.BufferData(BufferTarget.ArrayBuffer, 4 * 6 * 4, vquad, BufferUsageHint.StaticDraw);
			int _VAO = GL.GenVertexArray();
			GL.BindVertexArray(_VAO);
			GL.EnableVertexAttribArray(0);
			GL.VertexAttribPointer(0, 2, VertexAttribPointerType.Float, false, 4 * 4, 0);
			GL.EnableVertexAttribArray(1);
			GL.VertexAttribPointer(1, 2, VertexAttribPointerType.Float, false, 4 * 4, 2 * 4);
			GlobalBuffers.Add(win.Context, new(_VBO, _VAO, 0));
		}

		Tuple<int, int, int> tup = GlobalBuffers[win.Context];
		VBO = tup.Item1;
		VAO = tup.Item2;
		GlobalBuffers[win.Context] = new(VBO, VAO, tup.Item3 + 1);
		
		
		GL.BindBuffer(BufferTarget.ArrayBuffer, 0);
		GL.BindVertexArray(0);
		
		Loaded = true;
		Text = Text;
		Location = Location;
		if (Distance.X == 0 && Distance.Y == 0) ForceDistanceUpdate(Parent);
		CallWindowLoaded();
	}

	private void WindowOnMouseDown(MouseButtonEventArgs obj)
	{
		if (MouseInside && obj.Button == MouseButton.Button1)
		{
			bool subnothit = true;
			if (SubHitBoxes.Count > 0)
			{
				foreach (SubHitBox hb in SubHitBoxes)
				{
					if (hb.MouseInside)
					{
						hb.SendClick();
						subnothit = false;
					}
				}
			}
			if (subnothit) TrigerClick();
		}
		if (MouseInside && obj.Button == MouseButton.Button2 && ContextMenu is not null && Window!.HoveringControl == this) ContextMenu.ShowContext(Window!);
		if (obj.Button != MouseButton.Button2 && ContextMenu is not null && ContextMenu.Visible) ContextMenu.HideContext(Window!);
	}
}